package nplay.json.hornetq.client;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;

import nplay.common.app.MessagePriority;
import nplay.common.app.Queue;
import nplay.json.data.AJsonValue;
import nplay.json.data.CDRMessageValue;
import nplay.json.data.EmailOutMessageValue;
import nplay.json.data.MessageTicketValue;
import nplay.json.data.MmsInMessageValue;
import nplay.json.data.MmsOutMessageValue;
import nplay.json.data.SmsInMessageValue;
import nplay.json.data.SmsOutMessageValue;
import nplay.json.data.SnsInMessageValue;
import nplay.json.data.SnsOutMessageValue;
import nplay.json.jms.client.QueueMessageValue;
import nplay.json.rest.IQueueClient;

import org.hornetq.api.core.HornetQException;
import org.hornetq.api.core.TransportConfiguration;
import org.hornetq.api.core.client.ClientMessage;
import org.hornetq.api.core.client.ClientProducer;
import org.hornetq.api.core.client.ClientSession;
import org.hornetq.api.core.client.ClientSessionFactory;
import org.hornetq.api.core.client.HornetQClient;
import org.hornetq.api.core.client.ServerLocator;
import org.hornetq.core.remoting.impl.netty.NettyConnectorFactory;
import org.json.JSONException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * @author Alfred Wong
 * 
 */
public class QueueJsonHornetQClient implements IQueueClient {

	// Variables
	//************************************************************
	
	private static Logger logger = LoggerFactory.getLogger(QueueJsonHornetQClient.class);
	
	// Used to test HornetQ connection
	private final String TEST_QUEUE = "jms.queue.TEST_QUEUE";
	private final String TEST_MESSAGE_PARAM_NAME = "testmsg";
	private final String TEST_MESSAGE = "TEST CONNECTION ONLY";
	
	// Inject by setter method
	private List<String> hostPortList;
	private long expiration = 0;
	private int producerMaxRate = 500;
	private int consumerMaxRate = 500;
	// Recommended set total producer > 1 to improve produce TPS, 
	// one producer only able to support 100-150(max) TPS
	private int totalProducers = 1;
	// Recommended set re-poll producer interval 1, if set >1 produce TPS will drop
	private int repollProducerInterval = 1;
	// Consume time out must set >= 1
	private int consumeTimeout = 1;
	
	//Instance variables
	private ServerLocator serverLocator;
	private ClientSessionFactory clientSessionFactory;
	private AtomicBoolean isShutdown;
	private Map<Queue, HornetQQueueSession> hornetqQueueSessionList;
	
	// Default values
	private final int ACK_BATCH_SIZE = 0;
	
  	//************************************************************
	
	// Setter START **********************************************
	
	public void setHostPortList(List<String> hostPortList) {
		this.hostPortList = hostPortList;
	}
	
	public void setExpiration(long expiration) {
		this.expiration = expiration;
	}
	
	public void setProducerMaxRate(int producerMaxRate) {
		this.producerMaxRate = producerMaxRate;
	}

	public void setConsumerMaxRate(int consumerMaxRate) {
		this.consumerMaxRate = consumerMaxRate;
	}
	
	public void setTotalProducers(int totalProducers) {
		this.totalProducers = totalProducers;
	}

	public void setRepollProducerInterval(int repollProducerInterval) {
		this.repollProducerInterval = repollProducerInterval;
	}

	public void setConsumeTimeout(int consumeTimeout) {
		this.consumeTimeout = consumeTimeout;
	}
	
	// Setter END  ************************************************

	@Override
	public boolean send(SmsOutMessageValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority);
	}

	@Override
	public boolean send(SmsInMessageValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority);
	}

	@Override
	public boolean send(MmsOutMessageValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority);
	}

	@Override
	public boolean send(MmsInMessageValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority);
	}

	@Override
	public boolean send(EmailOutMessageValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority);
	}

	@Override
	public boolean send(SnsOutMessageValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority);
	}

	@Override
	public boolean send(SnsInMessageValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority);
	}

	@Override
	public boolean send(MessageTicketValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority);
	}

	@Override
	public boolean send(CDRMessageValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority); 
	}

	@Override
	public SmsOutMessageValue pollSmsOut(Queue queue) {
		return consumeTextMessage(queue);
	}

	@Override
	public SmsInMessageValue pollSmsIn(Queue queue) {
		return consumeTextMessage(queue);
	}

	@Override
	public MmsOutMessageValue pollMmsOut(Queue queue) {
		return consumeTextMessage(queue);
	}

	@Override
	public MmsInMessageValue pollMmsIn(Queue queue) {
		return consumeTextMessage(queue);
	}

	@Override
	public EmailOutMessageValue pollEmailOut(Queue queue) {
		return consumeTextMessage(queue);
	}

	@Override
	public SnsOutMessageValue pollSnsOut(Queue queue) {
		return consumeTextMessage(queue);
	}

	@Override
	public SnsInMessageValue pollSnsIn(Queue queue) {
		return consumeTextMessage(queue);
	}

	@Override
	public CDRMessageValue pollCDR(Queue queue) {
		return consumeTextMessage(queue);
	}

	@Override
	public boolean send(SmsOutMessageValue message, Queue queue) {
		return send(message, queue, MessagePriority.NORMAL); 
	}

	@Override
	public boolean send(SmsInMessageValue message, Queue queue) {
		return send(message, queue, MessagePriority.NORMAL); 
	}

	@Override
	public boolean send(MmsOutMessageValue message, Queue queue) {
		return send(message, queue, MessagePriority.NORMAL); 
	}

	@Override
	public boolean send(MmsInMessageValue message, Queue queue) {
		return send(message, queue, MessagePriority.NORMAL); 
	}

	@Override
	public boolean send(SnsOutMessageValue message, Queue queue) {
		return send(message, queue, MessagePriority.NORMAL); 
	}

	@Override
	public boolean send(SnsInMessageValue message, Queue queue) {
		return send(message, queue, MessagePriority.NORMAL); 
	}

	@Override
	public boolean send(MessageTicketValue message, Queue queue) {
		return send(message, queue, MessagePriority.NORMAL); 
	}

	@Override
	public boolean send(CDRMessageValue message) {
		return send(message, Queue.QUEUE_CDR, MessagePriority.NORMAL);
	}
	
	@Override
	public synchronized void shutdown() {
		isShutdown.set(true);
		closeClientResources();
		logger.info("QueueJsonHornetQClient shutdown successfully...");
	}
	
	/*
	 * IMPORTANT!!!!!!
	 * Method used to initialize variables, if using applicationContext.xml to inject it
	 * remember to set the "init-method" point to this method, if using POJO to create the client
	 * please ensure invoke this method, with this the client wont work properly
	 */
	public void initClient() throws HornetQClientException {
		if(!connectToHornetQServers()) {
			throw new HornetQClientException("Failed to connect to HornetQ server(s)");
		}
		isShutdown = new AtomicBoolean(false);
		hornetqQueueSessionList = new HashMap<Queue, HornetQQueueSession>();
	}
	
	/*
	 * Method used to create connection to HornetQ server
	 * 
	 * @return Result whether connection create successfully or failed
	 */
	private synchronized boolean connectToHornetQServers() {
//		// Create server locator
//		TransportConfiguration[] tconfigs = new TransportConfiguration[hostPortList.size()];
//		int currentIndex = 0;
//		for(String hostPort : hostPortList) {
//			tconfigs[currentIndex] = getTransportConfiguration(hostPort);
//			currentIndex++;
//		}
//		serverLocator = HornetQClient.createServerLocatorWithHA(tconfigs);
//		serverLocator.setProducerMaxRate(producerMaxRate);
//		serverLocator.setConsumerMaxRate(consumerMaxRate);
//		// Set block on durable send to true for both durable and non-durable message
//		// to have guarantee delivery to the HornetQ server
//		serverLocator.setBlockOnDurableSend(true);
//		serverLocator.setBlockOnNonDurableSend(true);
//		
//		// Create session factory
//		try {
//			clientSessionFactory = serverLocator.createSessionFactory();
//		} catch (Exception e) {
//			logger.error("Failed to create session factory", e);
//			return false;
//		}
//		return true;
		for(String hostPort : hostPortList) {
			// Create server locator
			// WONG: Need find tune this
			TransportConfiguration tconfig = getTransportConfiguration(hostPort);
			serverLocator = HornetQClient.createServerLocatorWithHA(tconfig);
			serverLocator.setProducerMaxRate(producerMaxRate);
			serverLocator.setConsumerMaxRate(consumerMaxRate);
			// Set block on durable send to true for both durable and non-durable message
			// to have guarantee delivery to the HornetQ server
			serverLocator.setBlockOnDurableSend(true);
			serverLocator.setBlockOnNonDurableSend(true);
//			serverLocator.setBlockOnAcknowledge(true);
			serverLocator.setAckBatchSize(ACK_BATCH_SIZE);
			serverLocator.setConnectionTTL(5000);
			serverLocator.setThreadPoolMaxSize(10);
			serverLocator.setUseGlobalPools(false);
			logger.info("Server locator created from {}", hostPort);
			
			// Create session factory
			try {
				clientSessionFactory = serverLocator.createSessionFactory();
				logger.info("Session factory created from {}", hostPort);
				return true;
			} catch (Exception e) {
				logger.error("Failed to create session factory from " + hostPort, e);
				// Close the failed host:port connection properly
				closeClientResources();
			}
		}
		return false;
	}
	
	/*
	 * Method used to create transport configuration based on host:port list provided
	 * 
	 * @param hostPort IMPORTANT!!! The string pattern must be "host:port" format 
	 * 
	 * @return Transport configuration object
	 */
	private TransportConfiguration getTransportConfiguration(String hostPort) {
		// EXCEPECT the format of each host-port pair values to be in "host:port" format
		String[] connectInfo = hostPort.split(":");
		String host = connectInfo[0];
		int port = Integer.parseInt(connectInfo[1]);
		Map<String, Object> params = new HashMap<String, Object>();
		params.put(org.hornetq.core.remoting.impl.netty.TransportConstants.HOST_PROP_NAME, host);
		params.put(org.hornetq.core.remoting.impl.netty.TransportConstants.PORT_PROP_NAME, port);
		TransportConfiguration tconfig = new TransportConfiguration(NettyConnectorFactory.class.getName(), params);
		return tconfig;
	}
	
	/*
	 * Method used to get HornetQ queue session from list
	 * 
	 * @param The requested HornetQ queue session must belong to this queue
	 */
	private HornetQQueueSession getHornetQQueueSession(Queue queue) {
		// Try get queue session from list
		HornetQQueueSession hornetqQueueSession = hornetqQueueSessionList.get(queue);
		// Check whether queue session succesfully get from list
		if(hornetqQueueSession == null) {
			// Sync the queue list
			synchronized (hornetqQueueSessionList) {
				// Check again because it might possible the queue session has been created by other thread
				hornetqQueueSession = hornetqQueueSessionList.get(queue);
				// If confirm queue session not exist, CREATE IT!
				if(hornetqQueueSession == null) {
					try {
						// Create queue session specifically for the queue
						hornetqQueueSession = new HornetQQueueSession(queue, clientSessionFactory, totalProducers, repollProducerInterval, consumeTimeout);
						// Store the created HornetQ queue session in list for shutdown, free up resources, etc purposes
						hornetqQueueSessionList.put(queue, hornetqQueueSession);
						return hornetqQueueSession;
					} catch (HornetQException e) {
						logger.error("Unable to get HornetQ queue session HornetQErrorCode["+e.getCode()+"]", e);	
					}
				}
			}
		}
		// Return existing queue session or newly created queue session back to caller
		return hornetqQueueSession;
	}
	
	/*
	 * Method used to produce text message to the queue and used the default expiration
	 * 
	 * @param message Actual JSON message
	 * @param queue Target queue
	 * @param priority Message priority
	 * 
	 * @return Success or failed produce message response
	 */
	public boolean produceTextMessage(AJsonValue message, Queue queue, MessagePriority priority) {
		return produceTextMessage(message, queue, priority, expiration);
    }
	
	/*
	 * Method used to produce text message to the queue
	 * HornetQ v2.2.14.Final info for expiration:
	 * HornetQ can assign an expiry address to a given queue so that when messages are expired, they
	 * are removed from the queue and sent to the expiry address. These "expired" messages can later
	 * be consumed from the expiry address for further inspection.
	 * 
	 * @param message Actual JSON message
	 * @param queue Target queue
	 * @param priority Message priority
	 * @param expiration The expiration for the produce the message in ms
	 * 
	 * @return Success or failed produce message response
	 */
	public boolean produceTextMessage(AJsonValue message, Queue queue, MessagePriority priority, long expiration) {
		// Check whether client has been shutdown before process the produce request
		if(checkIsShutdown()) {
			return false;
		}
		// Wrap message inside a new QueueMessageValue object instance
		QueueMessageValue queueMessage = new QueueMessageValue(message);
		do {
			// Get HornetQ session tight to the queue
			HornetQQueueSession session = getHornetQQueueSession(queue);
			// Process only queue session is valid and not null, null means something already happened, must check log for further details
			if(session != null) {
				try {
					// Used queue session to produce the message
					session.produceTextMessage(queueMessage.toString(), priority.getJmsPriority(), expiration);
					// If everything smooth, proceed with return true
					return true;
				} catch (HornetQException e) {
					// Check whether shutdown request by caller
					if(checkIsShutdown()) {
						return false;
					}
					// Try to reconnect HornetQ
					if(!reconnectToHornetQServers()) {
						break;
					}
				} catch (HornetQClientException e) {
					// Check whether shutdown request by caller
					if(checkIsShutdown()) {
						return false;
					}
					// Try to reconnect HornetQ
					if(!reconnectToHornetQServers()) {
						break;
					}
				}
				// WONG: Might need to add Thread.sleep to avoid heavy loop
				try {
					Thread.sleep(1);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			} else {
				logger.error("Unable get QueueSession for '{}'", queue);
				break;
			}
		} while (true);
		// Not able to produce message, return false response to caller
		return false;
    }
	
	/*
	 * Method used to consume message from particular queue stored in HornetQ server
	 * 
	 * @param queue The queue where message consume from
	 * 
	 * @return The converted JSON message
	 */
	private <V extends AJsonValue> V consumeTextMessage(Queue queue) {
		// Check whether client is shutdown before process the consume request
		if(checkIsShutdown()) {
			return null;
		}
		do {
			// Get HornetQ queue session for the particular queue
			HornetQQueueSession session = getHornetQQueueSession(queue);
			if(session != null) {
				String text = null;
				try {
					// Consume message through HornetQ queue session
					text = session.consumeTextMessage();
					// Process further if the text return is not null
					if(text != null) {
						try {
							// Wrap the text consumed inside a new QueueMessageValue object
							QueueMessageValue queueMessage = new QueueMessageValue(text);
							// Return message back to caller
							return queueMessage.getQueueMessage();
						} catch (JSONException e) {
							logger.error("Invalid Queue Message Value from [Queue, Message] {}", new Object[]{queue, text}, e);
						}
					} else {
						// Stop the loop due to no message consume from HornetQ server
						break;
					}
					// WONG: Might need to add Thread.sleep to avoid heavy loop
					try {
						Thread.sleep(1);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				} catch (HornetQException e) {
					// Check whether shutdown request by caller
					if(checkIsShutdown()) {
						return null;
					}
					// Try to reconnect HornetQ
					if(!reconnectToHornetQServers()) {
						break;
					}
				}
			} else {
				logger.error("Unable get QueueSession from '{}'", queue);
				break;
			}
		} while (true);
		// Return null response back to the caller
		return null;
	}

	/*
	 * Method used to reconnect the HornetQ server(s)
	 * This method also design for external caller to reconnect the connection manually
	 * reused the existing object instance
	 * 
	 * @return Result whether reconnect successful or failed
	 */
	public synchronized boolean reconnectToHornetQServers() {
		// Check whether shutdown request by caller
		if(checkIsShutdown()) {
			return false;
		}
		ClientSession testSession = null;
		ClientProducer testProducer = null;
		try {
			// Before close the resources and start reconnect at least test whether
			// connection back to normal to avoid unnecessary close and open resources, 
			// because other thread might already do the restart reconnect already prior 
			// to current thread
			testSession = clientSessionFactory.createSession(true, true, 0);
			testProducer = testSession.createProducer(TEST_QUEUE);
			ClientMessage testMessage = testSession.createMessage(true);
			testMessage.putStringProperty(TEST_MESSAGE_PARAM_NAME, TEST_MESSAGE);
			testProducer.send(testMessage);
			// If connection work fine then just return true
			return true;
		} catch (HornetQException e) {
			logger.error("Failed to test connection", e);
			// Closed the resources
			closeClientResources();
			// Start the actual reconnect and handle if reconnect not successful situation
			if(!connectToHornetQServers()) {
				closeClientResources();
				return false;
			}
		} finally {
			// Close the resources opened for the testing
			// Close test producer
			if(testProducer != null) {
				try {
					testProducer.close();
				} catch (HornetQException e) {
					logger.error("Failed to close test producer on " + TEST_QUEUE, e);
				}
			}
			// Close test session
			if(testSession != null) {
				try {
					testSession.close();
				} catch (HornetQException e) {
					logger.error("Failed to close test session on " + TEST_QUEUE, e);
				}
			}
		}
		return true;
	}
	
	/*
	 * Method used to close client resources
	 */
	private synchronized void closeClientResources() {
		// Close all HornetQ queue session(s)
		if(hornetqQueueSessionList != null && hornetqQueueSessionList.size() > 0) {
			for(Queue queue : hornetqQueueSessionList.keySet()) {
				HornetQQueueSession hornetqQueueSession = hornetqQueueSessionList.get(queue);
				hornetqQueueSession.closeQueueSession();
				hornetqQueueSessionList.remove(queue);
			}
		}
		// Close client session factory
		if(clientSessionFactory != null) {
			clientSessionFactory.close();
		}
		// Close server locator
		if(serverLocator != null) {
			serverLocator.close();
		}
	}
	
	/*
	 * Method used to check whether client has been shutdown
	 * 
	 * @return Result whether client has been shutdown (true) or not yet (false)
	 */
	private boolean checkIsShutdown() {
		// We used atomic boolean which is thread-safe
		if(isShutdown.get()) {
			logger.warn("Client has been shutdown..");
			return true;
		}
		return false;
	}

}
