package nplay.json.hornetq.client.test;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicInteger;

import nplay.common.app.Queue;
import nplay.json.data.EmailOutMessageValue;
import nplay.json.hornetq.client.HornetQClientException;
import nplay.json.hornetq.client.QueueJsonHornetQClient;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TestConsume {

	private static Logger logger = LoggerFactory.getLogger(TestConsume.class);
	
//	private static final int totalMessages = 500000;
	private static AtomicInteger consumedCounter = new AtomicInteger(0);
//	private static int stopCounter = 1;
	private static int retryLimit = 10;
	private static long retryInterval = 1000;
	private static AtomicInteger currentRetryCounter = new AtomicInteger(0);
	
	public static void main(String[] args) throws HornetQClientException, InterruptedException {
		List<String> hostPortList = new ArrayList<String>();
		
		//consume
		hostPortList.add("192.168.1.99:5445");
		hostPortList.add("192.168.1.95:5445");
		
		QueueJsonHornetQClient client = new QueueJsonHornetQClient();
		client.setHostPortList(hostPortList);
		client.setExpiration(0);
		client.setProducerMaxRate(500); // Take note this is for each ClientProducer not all of them
		client.setConsumerMaxRate(1000);
		client.setTotalProducers(3);
		client.setRepollProducerInterval(1);
		client.setConsumeTimeout(5000);
		client.initClient();
		
		int numConnection = 20;
		
		//CONSUME
		ExecutorService consumeExecutor = Executors.newFixedThreadPool(numConnection);
		for (int i = 0; i < numConnection; i++) {
			Runnable worker = new ConsumeTask(client);
			consumeExecutor.execute(worker);
		}
	}
	
	private static class ConsumeTask implements Runnable {
		QueueJsonHornetQClient client;
		public ConsumeTask(QueueJsonHornetQClient client) {
			this.client = client;
		}
		public void run() {	
			// Break the infinite loop yourself by force kill!
			while (true) {
				EmailOutMessageValue emailOut = client.pollEmailOut(Queue.QUEUE_EMAIL_OUT);
				if(emailOut != null) {
					consumedCounter.incrementAndGet();
					currentRetryCounter.set(0);
					logger.info("Current Consumed Counter: {}", consumedCounter.get());
//					if(consumedCounter.get() == totalMessages) {
//						// Another few more times, if exceed the consumed expected, means something WRONG!
//						for(int i=0; i<stopCounter; i++) {
//							emailOut = client.pollEmailOut(Queue.QUEUE_EMAIL_OUT);
//							if(emailOut != null) {
//								consumedCounter.incrementAndGet();
////								logger.info("Current Consumed Counter: {}", consumedCounter.get());
//							}
//						}
//						if(consumedCounter.get() == totalMessages) {
//							logger.info("### NORMAL!!! Total Consumed: {}", consumedCounter.get());
//							break;
//						} else {
//							logger.error("### WRONG!!! Total Consumed: {} (or more) VS Correct Total: {}", consumedCounter.get(), totalMessages);
//							break;
//						}
//					}
				} else {
					currentRetryCounter.incrementAndGet();
					if(currentRetryCounter.get() >= retryLimit) {
						client.shutdown();
						QueueJsonHornetQClientTest.main(new String[0]);
						System.exit(0);
					}
					logger.info("Current Retry Counter: {}", currentRetryCounter.get());
					try {
						Thread.sleep(retryInterval);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		}
	}
	
}
