package nplay.json.hornetq.client.test;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicInteger;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import nplay.common.app.MessagePriority;
import nplay.common.app.Queue;
import nplay.json.data.EmailOutMessageValue;
import nplay.json.hornetq.client.HornetQClientException;
import nplay.json.hornetq.client.QueueJsonHornetQClient;


public class TestProduce {

	private static Logger logger = LoggerFactory.getLogger(TestProduce.class);
	
	private static final int totalMessages = 100000;
	private static AtomicInteger messageId = new AtomicInteger(0);
	private static AtomicInteger producedCounter = new AtomicInteger(0);
	
	public static void main(String[] args) throws HornetQClientException, InterruptedException {
		List<String> hostPortList = new ArrayList<String>();
		
		// produce
		hostPortList.add("192.168.1.95:5445");
		hostPortList.add("192.168.1.99:5445");
		
		QueueJsonHornetQClient client = new QueueJsonHornetQClient();
		client.setHostPortList(hostPortList);
		client.setExpiration(0);
		client.setProducerMaxRate(500); // Take note this is for each ClientProducer not all of them
		client.setConsumerMaxRate(500);
		client.setTotalProducers(5);
		client.setRepollProducerInterval(1);
		client.setConsumeTimeout(1);
		client.initClient();
		
		int numConnection = 20;
		
		// PRODUCE
		int modulus = totalMessages % numConnection;
		if(modulus != 0) {
			logger.error("Invalid modulus: {}", modulus);
			System.exit(-1);
		}
		int loop = totalMessages / numConnection;
		
		ExecutorService produceExecutor = Executors.newFixedThreadPool(numConnection);
		for (int i = 0; i < numConnection; i++) {
			Runnable worker = new ProduceTask(client, loop);
			produceExecutor.execute(worker);
		}
		
		Thread.sleep(10000000);
	}
	
	private static class ProduceTask implements Runnable {
		QueueJsonHornetQClient client;
		int loop;
		public ProduceTask(QueueJsonHornetQClient client, int loop) {
			this.client = client;
			this.loop = loop;
		}
		public void run() {
			String fromAddress = "xxx@gmail.com";
			String toAddress = "xxx@gmail.com";
			String subject = "Email testing for HornetQ";
			int randomPriority = 1;
			
			EmailOutMessageValue produceMessage = new EmailOutMessageValue(fromAddress, toAddress, subject);
			
			for(int i=0; i<loop; i++) {
				produceMessage.setSubject(subject+" ["+messageId.incrementAndGet()+"]");
				boolean response = client.produceTextMessage(produceMessage, Queue.QUEUE_EMAIL_OUT, MessagePriority.getPriority(randomPriority));
				if(response) {
					producedCounter.incrementAndGet();
					if(producedCounter.get() == totalMessages) {
						logger.info("### Total Produced: {}", producedCounter.get());
						client.shutdown();
						System.exit(0);
					}
				}
				if(randomPriority >= 9) {
					randomPriority = 1;
				} else {
					randomPriority += 2;
				}
			}
		}
	}
	
}
